import time
import ee_manager.utils as utils
import ee_manager.logger as logger
import ee_manager.config as config

utils.set_timezone()

class BatchListener(object):
	# config
	DEFAULT_MIN_WAIT=60*5
	DEFAULT_MAX_WAIT=60*30
	NUM_TO_CHECK=20
	SLOW_DOWN=0.75
	PRINT=True
	# initialization attributes
	state=None
	tasks=None
	num_tasks=None
	max_wait=None
	min_wait=None
	callback_module=None

	
	#
	#	PUBLIC METHODS
	#
	def __init__(self,tasks=[],batch_data=None,callback=None,callback_module=None,min_wait=None,max_wait=None):
		self.tasks=tasks
		self.num_tasks=self.task_count()
		self.callback=callback
		self.callback_module=callback_module
		self.batch_data=batch_data
		self.min_wait=min_wait or self.DEFAULT_MIN_WAIT
		self.max_wait=max_wait or self.DEFAULT_MAX_WAIT
		self._initialize_attributes()
		self.state='INITIALIZED'


	def task_count(self):
		return len(self.tasks)


	def add_tasks(self,tasks_or_task,auto_listen=True):
		if type(tasks_or_task) is list: new_tasks=tasks_or_task
		else: new_tasks=[tasks_or_task]
		tasks = self.tasks or []
		utils.safe_exec(self._report,action='add tasks {}'.format(map(lambda t: "{}[{}]".format(t.description,t.id), new_tasks)))
		self.num_tasks=self.num_tasks+len(new_tasks)
		self.tasks=list(set(tasks+new_tasks))
		if auto_listen: self.listen() 


	def listen(self):
		if not self.is_listening():
			logger.safe_log(
				severity='INFO',
				text='BatchListener.listen',
				num_open_tasks=len(self.tasks),
				num_tasks=self.num_tasks,
				prior_state=self.state,
				state='LISTENING',
				trace_info='BatchListener.listen'
			)
			self.state='LISTENING'
			self._check_tasks()


	def is_listening(self):
		return self.state=='LISTENING'


	def cancel(self):
		self.state='CANCELLED'


	#
	#	INTERNAL METHODS
	#
	def _initialize_attributes(self):
		self.failed_tasks=[]
		self.completed_tasks=[]
		self.cancelled_tasks=[]
		self.unsubmitted_tasks=[]
		self.unclassified_tasks=[]
		# internal props
		self._check_count=0
		self._task=None


	def _check_tasks(self):
		if self.state=='LISTENING':
			task=self._next_task()
			if not task: self._finalize_batch()
			else:
				self._wait_for_task(task)
				for task in self._get_next_updated_tasks():
					if task.finished():
						task.finalize()
						self._close_task(task)
				self._check_tasks()


	def _next_task(self):
		return next(iter(self.tasks),None)


	def _get_next_updated_tasks(self):
		next_tasks=(self.tasks or [])[:self.NUM_TO_CHECK]
		for task in next_tasks:
			task.update_status()
			if task.finished(): 
				utils.safe_exec(self._report,action='task-{}[{}]: {}'.format(task.description,task.id,task.state))
		return next_tasks


	def _wait_for_task(self,task):
		max_wait=task.max_wait or self.max_wait
		min_wait=task.min_wait or self.min_wait
		wait_time=max((max_wait-(task.runtime()*self.SLOW_DOWN)),min_wait)
		utils.safe_exec(self._report,action='\twaiting... {}'.format(self._wait_time_fmt(wait_time)))
		utils.wait(wait_time,'BatchListener._wait_for_task')


	def _wait_time_fmt(self,wait_time):
		hours, remainder = divmod(wait_time, 3600)
		minutes, seconds = divmod(remainder, 60)
		return '{}:{}:{}'.format(hours, minutes, seconds)


	def _close_task(self,task):
		self.tasks=filter(lambda t: t.id!=task.id, self.tasks)
		logged_as=None
		severity=None
		if task.complete(): 
			self.completed_tasks.append(task)
			logged_as='completed'
		elif task.failed(): 
			self.failed_tasks.append(task)
			logged_as='failed'
			severity='ERROR'
		elif task.cancelled(): 
			self.cancelled_tasks.append(task)
			logged_as='cancelled'
			severity='WARNING'
		elif task.unsubmitted(): 
			self.unsubmitted_tasks.append(task)
			logged_as='unsubmitted'
			severity='ERROR'
		else: 
			self.unclassified_tasks.append(task)
			logged_as='unclassified'
			severity='ERROR'
		
		logger.safe_log(
			text='TASK CLOSED ({})'.format(logged_as),
			severity=severity,
			data=task.status,
			task_id=task.id,
			logged_as=logged_as,
			batch_listener={
				'num_open_tasks':len(self.tasks),
				'num_tasks':self.num_tasks,
				'state':self.state
			},
			trace_info='BatchListener._close_task'
		)
		self._reset()


	def _reset(self):
		self._task=None
		self._check_count=0


	def _finalize_batch(self):
		return_value=True
		self.state=self._get_final_state()
		if self.callback: 
			return_value=utils.exec_function(
				self.callback,
				self.callback_module,
				state=self.state,
				completed_tasks=self.completed_tasks,
				failed_tasks=self.failed_tasks,
				cancelled_tasks=self.cancelled_tasks,
				unclassified_tasks=self.unclassified_tasks)
		utils.safe_exec(self._report)
		return return_value


	def _get_final_state(self):
		if self.tasks: return 'INCOMPLETE'
		elif len(self.completed_tasks)==self.num_tasks: return 'COMPLETE'
		elif (len(self.completed_tasks)+len(self.cancelled_tasks))==self.num_tasks:
			return 'COMPLETE_WITH_CANCELLED'
		else:
			return 'COMPLETE_WITH_ERRORS'


	def _report(self,action=None):
		report_str="\t({})  {} [{}/{}]".format(
			utils.timenow(),
			action or self.state,
			self.task_count(),
			self.num_tasks)
		task=self._next_task()
		if task: report_str="{}: {}".format(report_str,task.id)
		if self.PRINT: print report_str

